@@ -28,7 +28,7 @@ COUNT - 1 messages are deleted.  A negative argument means
 the current message and the previous |COUNT| - 1 messages are
 deleted.
 
-When invoked on marked messages (via vm-next-command-uses-marks),
+When invoked on marked messages (via `vm-next-command-uses-marks'),
 only marked messages are deleted, other messages are ignored."
   (interactive "p")
   (if (interactive-p)
@@ -75,7 +75,7 @@ COUNT - 1 messages are undeleted.  A negative argument means
 the current message and the previous |COUNT| - 1 messages are
 deleted.
 
-When invoked on marked messages (via vm-next-command-uses-marks),
+When invoked on marked messages (via `vm-next-command-uses-marks'),
 only marked messages are undeleted, other messages are ignored."
   (interactive "p")
   (if (interactive-p)
@@ -115,7 +115,7 @@ the variables vm-subject-ignored-prefix and vm-subject-ignored-suffix.
 The optional prefix argument ARG specifies the direction to move
 if vm-move-after-killing is non-nil.  The default direction is
 forward.  A positive prefix argument means move forward, a
-negative argument means move backward, a zero argument means
+negative arugment means move backward, a zero argument means
 don't move at all."
   (interactive "p")
   (vm-follow-summary-cursor)
@@ -150,7 +150,56 @@ don't move at all."
 				      (eq vm-move-after-killing t))))
 	(vm-next-message arg t executing-kbd-macro))))
 
-(defun vm-expunge-folder (&optional shaddap)
+(defun vm-delete-duplicate-messages ()
+  "Delete duplicate messages in the current folder.
+This command works by computing an MD5 hash for the body ofeach
+non-deleted message in the folder and deleting messages that have
+a hash that has already been seen.  Messages that already deleted
+are never hashed, so VM will never delete the last copy of a
+message in a folder.  'Deleting' means flagging for deletion; you
+will have to expunge the messages with `vm-expunge-folder' to
+really get rid of them, as usual.
+
+When invoked on marked messages (via `vm-next-command-uses-marks'),
+only duplicate messages among the marked messages are deleted,
+unmarked messages are not hashed or considerd for deletion."
+  (interactive)
+  (vm-select-folder-buffer)
+  (vm-check-for-killed-summary)
+  (vm-error-if-folder-read-only)
+  (vm-error-if-folder-empty)
+  (let ((used-marks (eq last-command 'vm-next-command-uses-marks))
+	(mlist vm-message-list)
+	(table (make-vector 61 0))
+	hash
+	(del-count 0))
+    (if used-marks
+	(setq mlist (vm-select-marked-or-prefixed-messages 0)))
+    (save-excursion
+      (save-restriction
+	(widen)
+	(while mlist
+	  (if (vm-deleted-flag (car mlist))
+	      nil
+	    (setq hash (vm-md5-region (vm-text-of (car mlist))
+				      (vm-text-end-of (car mlist))))
+	    (if (intern-soft hash table)
+		(progn
+		  (vm-set-deleted-flag (car mlist) t)
+		  (vm-increment del-count))
+	      (intern hash table)))
+	  (setq mlist (cdr mlist)))))
+    (vm-display nil nil '(vm-delete-duplicate-messages)
+		(list this-command))
+    (if (zerop del-count)
+	(message "No messages deleted")
+      (message "%d message%s deleted"
+	       del-count
+	       (if (= 1 del-count) "" "s")))
+    (vm-update-summary-and-mode-line)))
+
+(defun vm-expunge-folder (&optional shaddap just-these-messages
+				    messages-to-expunge)
   "Expunge messages with the `deleted' attribute.
 For normal folders this means that the deleted messages are
 removed from the message list and the message contents are
@@ -161,7 +210,7 @@ message list.  If virtual mirroring is in effect for the virtual
 folder, the corresponding real messages are also removed from real
 message lists and the message contents are removed from real folders.
 
-When invoked on marked messages (via vm-next-command-uses-marks),
+When invoked on marked messages (via `vm-next-command-uses-marks'),
 only messages both marked and deleted are expunged, other messages are
 ignored."
   (interactive)
@@ -174,16 +223,19 @@ ignored."
   (vm-update-summary-and-mode-line)
   (if (not shaddap)
       (message "Expunging..."))
-  (let ((use-marks (eq last-command 'vm-next-command-uses-marks))
+  (let ((use-marks (and (eq last-command 'vm-next-command-uses-marks)
+			(null just-these-messages)))
 	(mp vm-message-list)
 	(virtual (eq major-mode 'vm-virtual-mode))
 	(buffers-altered (make-vector 29 0))
 	prev virtual-messages)
     (while mp
       (cond
-       ((and (vm-deleted-flag (car mp))
-	     (or (not use-marks)
-		 (vm-mark-of (car mp))))
+       ((if just-these-messages
+	    (memq (car mp) messages-to-expunge)
+	  (and (vm-deleted-flag (car mp))
+	       (or (not use-marks)
+		   (vm-mark-of (car mp)))))
 	;; remove the message from the thread tree.
 	(if vm-thread-obarray
 	    (vm-unthread-message (vm-real-message-of (car mp))))
@@ -263,7 +315,7 @@ ignored."
 	    ;; vm-clear-expunge-invalidated-undos uses this to recognize
 	    ;; expunged messages.
 	    (vm-set-deleted-flag-of (car mp) 'expunged)
-	    ;; disable summary any summary update that may have
+	    ;; disable any summary update that may have
 	    ;; already been scheduled.
 	    (vm-set-su-start-of (car mp) nil)
 	    (vm-set-buffer-modified-p t)
@@ -272,6 +324,24 @@ ignored."
 		(vm-attributes-of (vm-real-message-of (car mp))))
 	    (save-excursion
 	      (set-buffer (vm-buffer-of (vm-real-message-of (car mp))))
+	      (cond ((eq vm-folder-access-method 'pop)
+		     (setq vm-pop-messages-to-expunge
+			   (cons (vm-pop-uidl-of (vm-real-message-of (car mp)))
+				 vm-pop-messages-to-expunge)
+			   ;; Set this so that if Emacs crashes or
+			   ;; the user quites without saving, we
+			   ;; have a record of messages that were
+			   ;; retrieved and expunged locally.
+			   ;; When the user does M-x recover-file
+			   ;; we won't re-retrieve messages the
+			   ;; user has already deal with.
+			   vm-pop-retrieved-messages
+			   (cons (list (vm-pop-uidl-of
+					(vm-real-message-of (car mp)))
+				       (vm-folder-pop-maildrop-spec)
+				       'uidl)
+				 vm-pop-retrieved-messages))))
+	      (vm-increment vm-modification-counter)
 	      (vm-save-restriction
 	       (widen)
 	       (let ((buffer-read-only nil))
